from .generic import *


class Parser():

    def __init__(self, page: PAGE) -> None:
        self.page = page
        self.elements: ELEMENTS = None
        self.fields = {
            TABLES: ('tables',),
            ARRAY: ('array', 'last', 'first', 'merged',),
            ARRAYS: ('all',)}
        for TYPE in self.fields:
            for field in self.fields[TYPE]:
                self.__dict__ |= {field: TYPE()}

    def init(self) -> None:
        self.select()
        self.manage()
        self.export()

    def clean(self, string: str, mode: str = 'content') -> str:
        if mode == 'content':
            replacers = {'"':'', '\u200b': '', '\u2013': '-', '[edit]': '', '\n': ' '}
            string = re.sub(r'\[\d*\]', '', normalize('NFKD', string))
        elif mode == 'href':
            replacers = {'_':' ', '#': ''}
        for chars, replacer in replacers.items(): string = string.replace(chars, replacer)
        return string.strip()

    def prepare(self,
        elements: ELEMENTS | ELEMENT,
        tag: TAG = None,
        replacer: str = None
        ) -> None:
        if isinstance(elements, ELEMENT): elements = (elements,)
        for element in elements:
            if not isinstance(element, ELEMENT): continue
            for tag in element.find_all(tag.name, tag.attrs
                    ) if tag.attrs else element.find_all(tag.name):
                tag.replace_with(replacer) if replacer else tag.extract()

    def derive(self, tr: ELEMENT) -> tuple[ELEMENT | ELEMENTS]:
        ths: ELEMENTS = tr.find_all('th')
        tds: ELEMENTS = tr.find_all('td')
        th: ELEMENT = (ths[0], ths.clear())[0] if len(ths) == 1 else None
        td: ELEMENT = (tds[0], tds.clear())[0] if len(tds) == 1 else None
        ul: ELEMENT = td.find('ul') if td else None
        return ths, tds, th, td, ul

    def scope(self, element: ELEMENT) -> SCOPE:
        return SCOPE(
            row=int(re.split(r'(\s|[a-zA-Z])', row)[0]) if (row := element.get('rowspan')) else 1, 
            col=int(re.split(r'(\s|[a-zA-Z])', col)[0]) if (col := element.get('colspan')) else 1, 
            scope=scope if (scope := element.get('scope')) else None)

    def select(self, page: PAGE = None, tag: TAG = None) -> None:
        if not tag: tag = self.tag
        if not page: page = self.page
        if not page: return None
        self.heading: str = self.clean(page.html.find('h1').text)
        self.elements: ELEMENTS = page.content.find_all(
            tag.name, attrs=tag.attrs, recursive=tag.recursive
            ) if tag.attrs else page.content.find_all(tag.name, recursive=tag.recursive)

    def manage(self) -> None:
        for table in self.gather(self.elements):
            self.tables += (table,)

    def export(self) -> None:
        for order, table in enumerate(self.tables):
            self.array = table.data()
            self.merged |= self.array
            self.all += (self.array,)
        self.last = self.array
        if self.all: self.first = self.all[0]



#!/usr/bin/env python3
from optparse import OptionParser
from .driver import Wiki, OUTPUT


def run():
	parser = OptionParser()
	options = (
		('-u', '--url', 'url', None),
		('-q', '--query', 'query', None),
		('-l', '--lang', 'lang', 'English'),
		('-p', '--part', 'part', 'all'),
		('-i', '--item', 'item', 'all'),
		('-o', '--output', 'output', 'text'),)
	args = [dict(zip(('flag', 'long', 'dest', 'default'), option)) for option in options]
	for arg in args: parser.add_option(
		arg['flag'], arg['long'], dest=arg['dest'], default=arg['default'])
	options = parser.parse_args()[0]
	if options.output not in (formats := OUTPUT()._fields): 
		return print(f"Available output formats: {formats}.")
	output = Wiki(lang=options.lang).search(
	    url=options.url, query=options.query, part=options.part, item=options.item)
	print(getattr(output, options.output))

if __name__ == '__main__': run()



import re
from json import dumps
from urllib import request
from urllib.parse import quote, unquote
from unicodedata import normalize
from typing import Tuple, Dict, NamedTuple, TypedDict, TypeAlias, MutableMapping

from bs4 import BeautifulSoup, element


class TUPLE(Tuple[str]): pass
class DICT(Dict, MutableMapping[str, str | TUPLE]): pass

DATA: TypeAlias = str | TUPLE | DICT | dict[str, DICT]

HTML: TypeAlias = BeautifulSoup
ELEMENT: TypeAlias = element.Tag
ELEMENTS: TypeAlias = element.ResultSet | tuple[ELEMENT]

class PAGE(NamedTuple):
    url: str
    html: HTML
    content: ELEMENT = None

class TAG(NamedTuple):
    name: str | list[str] = None
    attrs: dict | None = None
    recursive: bool = True

class SCOPE(NamedTuple):
    row: int = 1
    col: int = 1
    scope: str = None

class CELL(NamedTuple):
    data: DATA | None
    scope: SCOPE = SCOPE()
    index: int = 1

class CELLS(Tuple[CELL]): pass

class ROW(NamedTuple):
    index: int = 1
    label: str | None = None
    cells: CELLS = CELLS()
    length = lambda self: len(self.cells)
    scope = lambda self: None if not self.cells else SCOPE(
        row=max(cell.scope.row for cell in self.cells),
        col=sum(cell.scope.col for cell in self.cells))
    data = lambda self: ARRAY({self.label: tuple([cell.data for cell in self.cells]
        ) if len(self.cells) != 1 else self.cells[0].data})

class ROWS(Tuple[ROW]): pass

class TABLE(NamedTuple):
    label: str | None = None
    rows: ROWS = ROWS()
    length = lambda self: len(self.rows)
    scope = lambda self: None if not self.rows else SCOPE(
        row=sum(row.scope().row for row in self.rows),
        col=max(row.scope().col for row in self.rows))
    data = lambda self: ARRAY({self.label:
        {label: data for row in self.rows for label, data in row.data().items()}})
    
class TABLES(Tuple[TABLE]): pass

class ARRAY(Dict, MutableMapping[str, DATA]): pass
class ARRAYS(Tuple[ARRAY]): pass

class OUTPUT(NamedTuple):
    dict: dict = None
    json: str = None
    text: str = None

FIELD: TypeAlias = CELL | ROW
FIELDS: TypeAlias = CELLS | ROWS
CONTEINER: TypeAlias = ROW | TABLE

def update(conteiner: CONTEINER, fields: FIELDS | FIELD) -> CONTEINER:
    if isinstance(fields, FIELD.__args__): fields = (fields,)
    for field in conteiner._fields:
        if fields and field.startswith(type(fields[0]).__name__.lower()): break
    return conteiner._replace(**{field: (*getattr(conteiner, field), *fields,)})

def merge(master: TABLE, slave: TABLE) -> TABLE:
    return master._replace(**{'rows': (*getattr(master, 'rows'), *getattr(slave, 'rows'),)})

def rename(conteiner: CONTEINER, label: str) -> CONTEINER:
    return conteiner._replace(**{'label': label})

def delete(conteiner: CONTEINER, field: str, index: int) -> CONTEINER:
    return conteiner._replace(**{field: ((fields := getattr(conteiner, field)),
        (fields[:index] + fields[index + 1:]))[1]})

def resize(cell: CELL, row: int = None, col: int = None) -> CELL:
    return cell._replace(**{'scope': SCOPE(
        row=row if row else cell.scope.row, col=col if col else cell.scope.col)})



from .generic import *
from .parser import Parser


class Infobox(Parser):

    def __init__(self, page: PAGE) -> None:
        Parser.__init__(self, page=page)
        self.tag = TAG(name='table', attrs={'class': ['infobox', 'infobox_v2']}, recursive=False)
        self.init()

    def extract(self, element: ELEMENT) -> TABLE:
        for inner in element.find_all(self.tag.name):
            inner.extract()
            yield self.parse(inner.find_all('tr'))

    def parse(self, elements: ELEMENTS) -> TABLE:
        table = TABLE(); row = ROW(); section = DICT(); order = 0
        stower = lambda row, section: (update(row, CELL(section)) if len(
            section) > 1 or len(section) == 1 and type(list(section.keys())[0]) != int else update(
            row, CELL(list(section.values()))) if len(section) == 1 and type(
            list(section.keys())[0]) == int else row) if section else row
        for tr in elements:
            if type(tr) != ELEMENT or not tr.text.strip(): continue
            label, data = '', ()
            ths, tds, th, td, ul = self.derive(tr)
            if th: 
                self.prepare(elements=th, tag=TAG(name='br'), replacer='&&')
                label = text[0] if ':' in th.text else ' '.join((text := th.text.split('&&')))
            if ul:
                data = [li.text.title() for li in ul.find_all('li')]
            elif td:
                if order <= 2 and (image := td.select_one('.image')):
                    section |= DICT({'Image': f"https:{image.find('img').get('src')}"})
                if td.find_all('br'):
                    self.prepare(elements=td, tag=TAG(name='br'), replacer='&&')
                    data = td.text.split('&&')
                elif next(td.children).name in ['div', 'span', 'time']:
                    data = [td.text.strip()]
                elif len((anchors := td.find_all('a', recursive=False))) > 1:
                    data = [anchor.text for anchor in anchors]
                else: data = [td.text.strip()]
            elif tds:
                data = [td.text for td in tds]
            label = self.clean(label)
            data = [clear for line in data if (clear := self.clean(line))]
            if not label and not data: continue
            if not label:
                if len(data) > 1 and ':' in data[0]: label = data.pop(0)
                if not label and len(section) > 0 and list(section.keys())[-1] == 'Image': label = 'Caption'
            if not label and len(data) == 1: label = data[0]; data = []
            if label and th and not td or not tds and not data:
                table = update(table, row) if (row := stower(row, section)).label else table
                row = ROW(label=label); section = DICT(); order = 0
                continue
            section |= DICT({label if label else order + 1: data[0] if len(data) == 1 else tuple(data)})
            order += 1
        table = update(table, row) if (row := stower(row, section)).label else table
        table = rename(table, table.rows[0].label) if table.length() else table
        return table

    def gather(self, elements: ELEMENTS) -> TABLE:
        for element in elements:
            nested = ROWS()
            self.prepare(elements=element, tag=TAG(name='span', attrs={'class': 'noprint'}))
            self.prepare(elements=element, tag=TAG(name='span', attrs={'style': 'display:none'}))
            self.prepare(elements=element, tag=TAG(name='td', attrs={'class': 'navigation-only'}))
            self.prepare(elements=element, tag=TAG(name='small'))
            self.prepare(elements=element, tag=TAG(attrs={'class': 'infobox-below'}))
            for order, inner in enumerate(self.extract(element=element)):
                if (title := inner.label) not in [row.label for row in nested]: label = title 
                else: label = f"{title} {order + 1}"
                nested += (ROW(label=label, cells=(CELL(data=inner.data()[title]),)),)
            if (caption := element.select_one('caption.infobox-title')):
                tr = HTML('<tr></tr>', 'html.parser'); caption.name = 'th'; tr.tr.insert(0, caption)
                table = self.parse([tr.find('tr')] + [tr for tr in element.find_all('tr')])
            else:
                table = self.parse(element.find_all('tr'))
            table = update(table, nested)
            yield table


class Paragraph(Parser):

    def __init__(self, page: PAGE) -> None:
        Parser.__init__(self, page=page)
        self.tag = TAG(name=['h2', 'h3', 'p'], recursive=False)
        self.init()

    def parse(self, elements: ELEMENTS) -> TABLE:
        row = None; header = 'Prologue'; headline = 'Prologue'; paragraphs = TUPLE()
        for order, element in enumerate(elements):
            if type(element) != ELEMENT: continue
            self.prepare(elements=element, tag=TAG(name='span', attrs={'class': 'noprint'}))
            self.prepare(elements=element, tag=TAG(name='small'))
            text = self.clean(element.text)
            if element.name == 'p':
                if text: paragraphs += (text,)
            elif element.name == 'h3':
                row = ROW(label=headline, cells=(CELL(paragraphs),))
                headline = text
                paragraphs = TUPLE()
            elif element.name == 'h2':
                if not row:
                    row = ROW(label=headline, cells=(CELL(paragraphs),))
                yield TABLE(label=header, rows=(row,))
                header = text

    def gather(self, elements: ELEMENTS) -> TABLE:
        for table in self.parse(elements):
            if table.rows[0].cells[0].data: yield table


class Table(Parser):

    def __init__(self, page: PAGE) -> None:
        Parser.__init__(self, page=page)
        self.tag = TAG(name='table', attrs={'class': 'wikitable'}, recursive=True)
        self.init()

    def parse(self, elements: ELEMENTS) -> TABLE:
        table = TABLE(); headers = ROWS(); previous = None
        packer = lambda element, order: CELL(
            data=self.clean(element.text), scope=self.scope(element))
        for line, tr in enumerate(elements):
            if type(tr) != ELEMENT: continue
            ths, tds, th, td, ul = self.derive(tr)
            cells = []
            for rows in (ths, (th,), tds, (td,)):
                if not rows or type(rows[0]) != ELEMENT: continue
                cells += [packer(row, order) for order, row in enumerate(rows)]
            if previous:
                for order, cell in enumerate(previous):
                    if (colspan := cell.scope.col) > 1:
                        for i in range(1, colspan):
                            previous.insert(order + i, CELL(data='', scope=SCOPE(col=0)))
                    if (rowspan := cell.scope.row) > 1:
                        mutual = resize(cell, row=rowspan - 1)
                        if order == 0:
                            label = re.split(r'(\s\[\d\])', mutual.data)[0] + f" [{mutual.index + 1}]" 
                            mutual = CELL(data=label, scope=mutual.scope, index=mutual.index + 1)
                        previous[order] = mutual
                        cells.insert(order, mutual)
                    else:
                        if len(cells) - 1 >= order: previous[order] = cells[order]
            else: previous = cells
            if len(cells) > 1 and not cells[0].data: cells.pop(0)
            if not tds and not td: row = ROW(label=f"Headers {line + 1}", cells=CELLS(cells))
            else: row = ROW(label=cells[0].data, cells=CELLS(cells[1:]))
            if len(row.cells) > 0 and row.cells[0].data: table = update(table, row)
        yield table

    def gather(self, elements: ELEMENTS) -> TABLE:
        for element in elements:
            root = TABLE(label=self.clean(title.text) if (
                title := element.find_previous('h3')) else 'No header')
            for table in self.parse(element.find_all('tr')):
                yield merge(root, table)


class List(Parser):

    def __init__(self, page: PAGE) -> None:
        Parser.__init__(self, page=page)
        self.tag = TAG(name='ul', recursive=False)
        self.init() 

    def parse(self, elements: ELEMENTS) -> TABLE:
        table = TABLE()
        for order, li in enumerate(elements):
            table = update(table, ROW(label=order + 1, cells=(CELL(self.clean(li.text)),)))
        yield table

    def gather(self, elements: ELEMENTS) -> TABLE:
        for element in elements:
            header = self.clean(title.text) if (title := element.find_previous('h2')) else 'No header'
            for table in self.parse(element.find_all('li', recursive=False)):
                yield rename(table, header)
                

class Thumb(Parser):

    def __init__(self, page: PAGE) -> None:
        Parser.__init__(self, page=page)
        self.tag = TAG(name='div', attrs={'class': 'thumb'}, recursive=False)
        self.init()

    def parse(self, elements: ELEMENTS) -> TABLE:
        for order, element in enumerate(elements):
            anchor: ELEMENT = element.find('a', {'class': 'image'})
            image = f"https:{anchor.find('img').get('src')}" if anchor else None
            header = self.clean(string=anchor.get(
                'href').split(':')[1].split('.')[0], mode='href') if anchor else order
            caption = self.clean(caption.text) if (caption := element.select_one('.thumbcaption')
                ) else 'No caption'
            yield TABLE(label=header, rows=(ROW(label=caption, cells=(CELL(image),)),))

    def gather(self, elements: ELEMENTS) -> TABLE:
        for table in self.parse(elements):
            yield table


class Toc(Parser):

    def __init__(self, page: PAGE) -> None:
        Parser.__init__(self, page=page)
        self.tag = TAG(name='div', attrs={'id': 'toc'}, recursive=True)
        self.init()

    def parse(self, elements: ELEMENTS) -> TABLE:
        table = TABLE()
        packer = lambda li, order: ROW(
            label=number.text if (number := li.select_one('.tocnumber')) else str(order + 1),
            cells=(CELL(text.text if (text := li.select_one('.toctext')) else 'No text'),))
        for order, li in enumerate(elements):
            table = update(table, packer(li, order))
            if (nested := li.find_all('li')):
                for inner in nested:
                    table = update(table, packer(inner, order))
        yield table

    def gather(self, elements: ELEMENTS) -> TABLE:
        for element in elements:
            header = title.text if (title := element.select_one('.toctitle')) else 'No title'
            if (ul := element.find('ul')):
                for table in self.parse(ul.find_all('li', recursive=False)):
                    yield rename(table, header)



languages = {
  'en': ['English'],
  'uk': ['Ukrainian', 'Українська'],
  'ru': ['Russian', 'Русский'],
  'pl': ['Polish', 'Polski'],
  'de': ['German', 'Deutsch'],
  'nl': ['Nederlands', 'Dutch'],
  'sv': ['Swedish', 'Svenska'],
  'es': ['Spanish', 'Español'],
  'fr': ['French', 'Français'],
  'it': ['Italian', 'Italiano'],
  'ja': ['Japanese'],
  'zh': ['Chainese'],
  'ceb': ['Cebuano', 'Bisaya'],
}

linguist = lambda lang: lang if (lang := lang.lower()) in languages else list(
    languages.keys())[list(languages.values()).index(value[0])] if (value := [
    value for value in languages.values() if lang.title() in value]) else 'en'


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wiki_fetch']

package_data = \
{'': ['*']}

install_requires = \
['bs4>=0.0.1,<0.0.2']

entry_points = \
{'console_scripts': ['wiki-fetch = wiki_fetch.__main__:run']}

setup_kwargs = {
    'name': 'wiki-fetch',
    'version': '0.0.9',
    'description': 'Parser for Wikipedia.org',
    'long_description': '# wiki-fetch\n\n[![PyPI](https://img.shields.io/pypi/v/wiki-fetch)](https://github.com/d3z-the-dev/wiki-fetch/releases/)\n[![Status](https://img.shields.io/pypi/status/wiki-fetch)](https://pypi.org/project/wiki-fetch/)\n[![PyPI Downloads](https://img.shields.io/pypi/dm/wiki-fetch)](https://pypi.org/project/wiki-fetch/)\n[![Python Version](https://img.shields.io/pypi/pyversions/wiki-fetch?color=%23244E71)](https://pypi.org/project/wiki-fetch/)\n[![License](https://img.shields.io/pypi/l/wiki-fetch?color=272727)](https://en.wikipedia.org/wiki/Apache_License#Apache_License_2.0)\n[![Issues](https://img.shields.io/github/issues/d3z-the-dev/wiki-fetch)](https://github.com/d3z-the-dev/wiki-fetch/issues)\n\n## Installation\n\n- PyPI\n\n```bash\npip install wiki-fetch\n```\n\n- Source\n\n```bash\ngit clone git@github.com:d3z-the-dev/wiki-fetch.git\ncd wiki-fetch && poetry build\npip install ./dist/*.whl\n```\n\n## Usage\n\n### CLI\n\n| Option           | Flag | Long       | Default | Example                                   |\n| ---------------- | ---- | ---------- | ------- | ----------------------------------------- |\n| Wiki\'s page link | `-u` | `--url`    | None    | <https://en.wikipedia.org/wiki/The_Doors> |\n| Search query     | `-q` | `--query`  | None    | The Doors (band)                          |\n| Page language    | `-l` | `--lang`   | English | English                                   |\n| Part of the page | `-p` | `--part`   | all     | infobox                                   |\n| Parts by order   | `-i` | `--item`   | all     | first                                     |\n| Output format    | `-o` | `--output` | text    | text                                      |\n\n```bash\nwiki-fetch -q \'The Doors (band)\' -p infobox -i first\n```\n\n<details>\n<summary>output</summary>\n\n```yaml\nInfobox: \n    The Doors: \n        The Doors: \n            Image: https://upload.wikimedia.org/wikipedia/commons/thumb/6/69/The_Doors_1968.JPG/250px-The_Doors_1968.JPG\n            Caption: The Doors in 1966: Morrison (left), Densmore (centre), Krieger (right) and Manzarek (seated)\n        Background information: \n            Origin: Los Angeles, California, U.S.\n            Genres: \n                Psychedelic Rock\n                Blues Rock\n                Acid Rock\n            Years active: \n                1965-1973\n                1978\n            Labels: \n                Elektra\n                Rhino\n            Spinoffs: \n                The Psychedelic Rangers\n                Butts Band\n                Nite City\n                Manzarek-Krieger\n            Spinoff of: Rick & the Ravens\n            Past members: \n                Jim Morrison\n                Ray Manzarek\n                Robby Krieger\n                John Densmore\n            Website: thedoors.com\nURL: https://en.wikipedia.org/?search=The Doors (Band)\n```\n</details>\n\n### Python\n\n| Argument | Values                                                         | Description                     |\n| -------- | -------------------------------------------------------------- | ------------------------------- |\n| url      | `str`                                                          | Any Wiki\'s page URL             |\n| query    | `str`                                                          | Any query string                |\n| lang     | `str`                                                          | Any of available languages      |\n| part     | `infobox`, `paragraph`, `table`, `list`, `thumb`, `toc`, `all` | Specify page part               |\n| item     | `first`, `last`, `all`                                         | Specify the order of the part   |\n\n```python\nfrom wiki_fetch.driver import Wiki\n\noutput = Wiki().search(query=\'The Doors (band)\', part=\'infobox\', item=\'first\')\nprint(output.json)\n```\n\n<details>\n<summary>output</summary>\n\n```json\n{\n    "Infobox": [\n        {\n            "The Doors": {\n                "The Doors": {\n                    "Image": "https://upload.wikimedia.org/wikipedia/commons/thumb/6/69/The_Doors_1968.JPG/250px-The_Doors_1968.JPG",\n                    "Caption": "The Doors in 1966: Morrison (left), Densmore (centre), Krieger (right) and Manzarek (seated)"\n                },\n                "Background information": {\n                    "Origin": "Los Angeles, California, U.S.",\n                    "Genres": [\n                        "Psychedelic Rock",\n                        "Blues Rock",\n                        "Acid Rock"\n                    ],\n                    "Years active": [\n                        "1965-1973",\n                        "1978"\n                    ],\n                    "Labels": [\n                        "Elektra",\n                        "Rhino"\n                    ],\n                    "Spinoffs": [\n                        "The Psychedelic Rangers",\n                        "Butts Band",\n                        "Nite City",\n                        "Manzarek-Krieger"\n                    ],\n                    "Spinoff of": "Rick & the Ravens",\n                    "Past members": [\n                        "Jim Morrison",\n                        "Ray Manzarek",\n                        "Robby Krieger",\n                        "John Densmore"\n                    ],\n                    "Website": "thedoors.com"\n                }\n            }\n        }\n    ],\n    "URL": "https://en.wikipedia.org/?search=The Doors (Band)"\n}\n```\n</details>\n',
    'author': 'd3z',
    'author_email': 'd3z.the.dev@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/d3z-the-dev/wiki-fetch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from .generic import *
from . import parts
from . import stuff


class Wiki():

    def __init__(self, lang: str = 'English') -> None:
        self.wiki = f"https://{stuff.linguist(lang)}.wikipedia.org/"
        self.headers = {
            'User-Agent': f"Mozilla/5.0 (X11; U; Linux x86_64) Gecko/081222 Firefox/108.0"}

    def get(self, url: str) -> str | None:
        response: http.client.HTTPResponse = request.urlopen(
            request.Request(url, headers=self.headers)) if url else None
        return response.read().decode('utf-8') if response else None

    def extract(self, response: str | None) -> HTML | None:
        return HTML(response, 'html.parser') if response else None

    def capture(self, url: str) -> PAGE:
        html = self.extract(self.get(url))
        if not html: return None
        contents: ELEMENT | None = html.find('div', {'id': 'bodyContent'})
        resuslts: ELEMENT | None = html.find('div', {'class': 'searchresults'})
        catlinks: ELEMENT | None = html.find('div', {'id': 'catlinks'})
        if resuslts:
            results = resuslts.find('ul'); resuslt = resuslts.find('li') if resuslts else None
            url = self.wiki + resuslt.find('a').get('href') if resuslt else None        
        elif catlinks and 'Disambiguation pages' in [li.text for li in catlinks.find('ul')]:
            href = [a.get('href') for a in contents.find_all('a') if a.get('href').startswith('/wiki/')][0]
            url = f"{self.wiki}{href}"
        html = self.extract(self.get(url)) if resuslts or catlinks else html
        body = html.select_one('#bodyContent')
        content = body.select_one('.mw-parser-output') if body else None
        return PAGE(url=url, html=html, content=content)

    def search(self, query: str = None, url: str = None, part: str = None, item: str = None) -> dict:
        self.output = dict()
        getter = lambda part, item: parts.__dict__[part.title()](page=self.page).__dict__[item]
        texter = lambda json: '\n'.join([line[4:] for line in re.sub(
            r'\",|\"|\},|\{|\}|\],|\[|\]', '', json).split('\n') if line.strip()])
        if not url and not query: return OUTPUT(text='There is no input data (query string or URL).')
        if query: url = f"{self.wiki}?search={quote(query.title())}"
        self.page = self.capture(url)
        Parts = ('Infobox', 'Paragraph', 'Table', 'List', 'Thumb', 'Toc',)
        [self.output.update({part: getter(part, item)}) for part in Parts
            ] if part == 'all' else self.output.update({part.title(): getter(part, item)})
        self.output |= {'URL': unquote(self.page.url)}
        return OUTPUT(
            dict=self.output,
            json=(json := dumps(self.output, indent=4, ensure_ascii=False, allow_nan=True)),
            text=texter(json))
